import * as http from 'http';
import * as needle from 'needle';
// import tar = require('tar');
// we have to use ES6 imports as it's providing correct types for chalk.
import chalk from 'chalk';
import path = require('path');
import fs = require('fs');
import urljoin = require('url-join');
import sumchecker = require('sumchecker');
import zlib = require('zlib');
import rimraf = require('rimraf');

// eslint-disable-next-line @typescript-eslint/no-var-requires
const config = require('libnpmconfig');

export class PactEnvironment {
  public get cwd(): string {
    return path.resolve(__dirname, '..');
  }

  public isWindows(platform: string = process.platform): boolean {
    return platform === 'win32';
  }
}

const pactEnvironment = new PactEnvironment()



// Sets the needle default for all calls through npm environment variables for proxy
const environmentProxy =
  process.env['npm_config_https_proxy'] || process.env['npm_config_proxy'];

if (environmentProxy) {
  needle.defaults({
    proxy: environmentProxy,
  });
}
// https://github.com/pact-foundation/pact-reference/releases/download/libpact_ffi-v0.4.3/libpact_ffi-osx-aarch64-apple-darwin.dylib.gz
export const PACT_FFI_VERSION = '0.4.3';
const PACT_DEFAULT_LOCATION = `https://github.com/pact-foundation/pact-reference/releases/download/libpact_ffi-v${PACT_FFI_VERSION}/`;
const HTTP_REGEX = /^http(s?):\/\//;

function join(...paths: string[]): string {
  return HTTP_REGEX.test(paths[0]) ? urljoin(...paths) : path.join(...paths);
}

function throwError(msg: string): never {
  throw new Error(chalk.red(`Error while installing ffi binary: ${msg}`));
}

function makeError(msg: string): Error {
  return new Error(chalk.red(`Error while locating pact ffi binary: ${msg}`));
}

function getBinaryLocation(
  location: string,
  basePath: string
): string | undefined {
  // Check if location is valid and is a string
  if (!location || location.length === 0) {
    return undefined;
  }

  // Check if it's a URL, if not, try to resolve the path to work with either absolute or relative paths
  return HTTP_REGEX.test(location)
    ? location
    : path.resolve(basePath, location);
}
function findPackageConfig(location: string, tries = 10): PackageConfig {
  if (tries === 0) {
    return {};
  }
  const packagePath = path.resolve(location, 'package.json');
  if (fs.existsSync(packagePath)) {
    // eslint-disable-next-line @typescript-eslint/no-var-requires, prefer-destructuring, import/no-dynamic-require, global-require
    const pkgConfig = require(packagePath).config;
    if (pkgConfig && (pkgConfig.pact_binary_location || pkgConfig.pact_do_not_track)) {
      return {
        binaryLocation: getBinaryLocation(
          pkgConfig.pact_binary_location,
          location
        ),
        doNotTrack: pkgConfig.pact_do_not_track,
      };
    }
  }

  return findPackageConfig(path.resolve(location, '..'), tries - 1);
}

export function createConfig(location: string = process.cwd()): Config {
  const packageConfig = findPackageConfig(location);
  const PACT_BINARY_LOCATION =
    packageConfig.binaryLocation || PACT_DEFAULT_LOCATION;
  const CHECKSUM_SUFFIX = '.sha256';
  const PREFIX = 'lib'
  return {
    doNotTrack:
      packageConfig.doNotTrack ||
      process.env['PACT_DO_NOT_TRACK'] !== undefined ||
      false,
    binaries: [
      {
        platform: 'win32',
        binaryFilename: 'pact_ffi.dll',
        binary: `pact_ffi-windows-x86_64.dll.gz`,
        binaryChecksum: `pact_ffi-windows-x86_64.dll.gz${CHECKSUM_SUFFIX}`,
        folderName: `winx64`,
        downloadLocation: PACT_BINARY_LOCATION,
      },
      {
        platform: 'darwin',
        arch: 'x64',
        binaryFilename: 'libpact_ffi.dylib',
        binary: `${PREFIX}pact_ffi-osx-x86_64.dylib.gz`,
        binaryChecksum: `${PREFIX}pact_ffi-osx-x86_64.dylib.gz${CHECKSUM_SUFFIX}`,
        folderName: `osxx64`,
        downloadLocation: PACT_BINARY_LOCATION,
      },
      {
        platform: 'darwin',
        arch: 'arm64',
        binaryFilename: 'libpact_ffi.dylib',
        binary: `${PREFIX}pact_ffi-osx-aarch64-apple-darwin.dylib.gz`,
        binaryChecksum: `${PREFIX}pact_ffi-osx-aarch64-apple-darwin.dylib.gz${CHECKSUM_SUFFIX}`,
        folderName: `osxaarch64`,
        downloadLocation: PACT_BINARY_LOCATION,
      },
      {
        platform: 'linux',
        arch: 'x64',
        binaryFilename: 'libpact_ffi.so',
        binary: `${PREFIX}pact_ffi-linux-x86_64.so.gz`,
        binaryChecksum: `${PREFIX}pact_ffi-linux-x86_64.so.gz${CHECKSUM_SUFFIX}`,
        folderName: `linuxx64`,
        downloadLocation: PACT_BINARY_LOCATION,
      },
      {
        platform: 'linux',
        arch: 'arm64',
        binaryFilename: 'libpact_ffi.so',
        binary: `${PREFIX}pact_ffi-linux-aarch64.so.gz`,
        binaryChecksum: `${PREFIX}pact_ffi-linux-aarch64.so.gz${CHECKSUM_SUFFIX}`,
        folderName: `linuxaarch64`,
        downloadLocation: PACT_BINARY_LOCATION,
      },
    ],
  };
}

const CONFIG = createConfig();

const CIs = [
  'CI',
  'CONTINUOUS_INTEGRATION',
  'ABSTRUSE_BUILD_DIR',
  'APPVEYOR',
  'BUDDY_WORKSPACE_URL',
  'BUILDKITE',
  'CF_BUILD_URL',
  'CIRCLECI',
  'CODEBUILD_BUILD_ARN',
  'CONCOURSE_URL',
  'DRONE',
  'GITLAB_CI',
  'GO_SERVER_URL',
  'JENKINS_URL',
  'PROBO_ENVIRONMENT',
  'SEMAPHORE',
  'SHIPPABLE',
  'TDDIUM',
  'TEAMCITY_VERSION',
  'TF_BUILD',
  'TRAVIS',
  'WERCKER_ROOT',
];

function downloadFileRetry(
  url: string,
  filepath: string,
  retry = 3
): Promise<unknown> {
  return new Promise(
    (
      resolve: (unused?: unknown) => void,
      reject: (e: string) => void
    ): void => {
      let len = 0;
      let downloaded = 0;
      let time = Date.now();
      let ca = config.read().cafile;
      if (ca) {
        ca = fs.readFileSync(ca);
      }
      needle
        .get(url, {
          follow_max: 5,
          headers: {
            'User-Agent': 'https://github.com/pact-foundation/pact-core',
          },
          rejectUnauthorized: config.read()['strict-ssl'] || false,
          ca,
        })
        .on('error', (e: string) => reject(e))
        .on(
          'response',
          // eslint-disable-next-line no-return-assign
          (res: http.IncomingMessage): number =>
            (len = parseInt(res.headers['content-length'] as string, 10))
        )
        .on('data', (chunk: string[]) => {
          downloaded += chunk.length;
          // Only show download progress every second
          const now = Date.now();
          if (now - time > 1000) {
            time = now;
            console.log(
              chalk.gray(
                `Downloaded ${((100 * downloaded) / len).toFixed(2)}%...`
              )
            );
          }
        })
        .pipe(fs.createWriteStream(filepath))
        .on('finish', resolve);
    }
  ).catch((e: string) =>
    // eslint-disable-next-line no-plusplus, no-param-reassign
    retry-- === 0 ? throwError(e) : downloadFileRetry(url, filepath, retry)
  );
}

function download(data: Data): Promise<Data> {
  console.log(
    chalk.gray(`Installing Pact Standalone Binary for ${data.platform}.`)
  );
  return new Promise(
    (resolve: (f: Data) => void, reject: (e: string) => void): void => {
      if (fs.existsSync(path.resolve(data.platformFolderPath))) {
        console.log(chalk.yellow('Binary already downloaded, skipping...'));
        resolve({ ...data, binaryAlreadyDownloaded: true });
        return;
      }
      fs.mkdirSync(data.platformFolderPath as string);
      console.log(
        chalk.yellow(
          `Downloading Pact Standalone Binary v${PACT_FFI_VERSION} for platform ${data.platform} from ${data.binaryDownloadPath}`
        )
      );

      // Track downloads through Google Analytics unless testing or don't want to be tracked
      if (!CONFIG.doNotTrack) {
        console.log(
          chalk.gray(
            'Please note: we are tracking this download anonymously to gather important usage statistics. ' +
              "To disable tracking, set 'pact_do_not_track: true' in your package.json 'config' section."
          )
        );
        // Trying to find all environment variables of all possible CI services to get more accurate stats
        // but it's still not 100% since not all systems have unique environment variables for their CI server
        const isCI = CIs.some(key => process.env[key] !== undefined);
        needle
          .post('https://www.google-analytics.com/collect', {
            form: {
              v: 1,
              tid: 'UA-117778936-1', // Tracking ID / Property ID.
              cid: Math.round(2147483647 * Math.random()).toString(), // Anonymous Client ID.
              t: 'screenview', // Screenview hit type.
              an: 'pact-ffi-install', // App name.
              // eslint-disable-next-line @typescript-eslint/no-var-requires, global-require
              av: require('../package.json').version, // App version.
              aid: 'pact-core', // App Id.
              aiid: `ffi-${PACT_FFI_VERSION}`, // App Installer Id.
              cd: `download-node-${data.platform}-${isCI ? 'ci' : 'user'}`,
              aip: true, // Anonymise IP address
            },
          })
          .on('error', () => {
            /* Ignore all errors */
          });
      }

      // Get archive of release
      // If URL, download via HTTP
      const basename = path.basename(data.filepath);
      const compressedFilePath = data.platformFolderPath ? path.join(data.platformFolderPath,basename) : basename
      if (HTTP_REGEX.test(data.binaryDownloadPath)) {
        downloadFileRetry(data.binaryDownloadPath, compressedFilePath).then(
          () => {
            console.log(
              chalk.green(`Finished downloading binary to ${data.filepath}`)
            );
            resolve(data);
          },
          (e: string) =>
            // eslint-disable-next-line prefer-promise-reject-errors
            reject(
              `Error downloading binary from ${data.binaryDownloadPath}: ${e}`
            )
        );
      } else if (fs.existsSync(data.binaryDownloadPath)) {
        // Or else it might be a local file, try to copy it over to the correct directory
        fs.createReadStream(data.binaryDownloadPath)
          .on('error', (e: string) =>
            // eslint-disable-next-line prefer-promise-reject-errors
            reject(
              `Error reading the file at '${data.binaryDownloadPath}': ${e}`
            )
          )
          .pipe(
            fs
              .createWriteStream(compressedFilePath)
              .on('error', (e: string) =>
                // eslint-disable-next-line prefer-promise-reject-errors
                reject(`Error writing the file to '${data.filepath}': ${e}`)
              )
              .on('close', () => resolve(data))
          );
      } else {
        // eslint-disable-next-line prefer-promise-reject-errors
        reject(
          `Could not get binary from '${data.binaryDownloadPath}' as it's not a URL and does not exist at the path specified.`
        );
      }
    }
  );
}

function extract(data: Data): Promise<Data> {

  const basename = path.basename(data.filepath);
  const compressedFilePath = data.platformFolderPath ? path.join(data.platformFolderPath,basename) : basename
  const binaryLocation = data.platformFolderPath ? path.join(data.platformFolderPath,data.binaryFilename): data.binaryFilename

  // If platform folder exists, binary already installed, skip to next step.
  if (fs.existsSync(binaryLocation as string)) {
    return Promise.resolve({ ...data, binaryAlreadyInstalled: true });
  }

  // Make sure checksum is available
  if (!fs.existsSync(data.checksumFilepath)) {
    throwError(`Checksum file missing from standalone directory. Aborting.`);
  }


  console.log(chalk.yellow(`Extracting binary from ${data.filepath}.`));

  // Validate checksum to make sure it's the correct binary

  return (
    sumchecker('sha256', data.checksumFilepath, data.platformFolderPath ? data.platformFolderPath : __dirname, basename)
      .then(
        () => console.log(chalk.green(`Checksum passed for '${basename}'.`)),
        () =>
          throwError(
            `Checksum rejected for file '${basename}' with checksum ${path.basename(
              data.checksumFilepath
            )}`
          )
      )
      // Extract files into their platform folder
      .then(() =>
          zlib.gunzip(fs.readFileSync(compressedFilePath), (err, buffer) => {
            fs.writeFileSync(binaryLocation,buffer)
              })
          
      )
      .then(() => {
        rimraf(compressedFilePath, () => console.log(`DELETED ./${compressedFilePath}/`));
        console.log(chalk.green('Extraction done.'));
      })
      .then(() => {
        console.log(
          `\n\n${ 
            chalk.bgYellow(
              chalk.black('### If you') +
                chalk.red(' ❤ ') +
                chalk.black('Pact and want to support us, please donate here:')
            ) 
            }${chalk.blue(' http://donate.pact.io/node') 
            }\n\n`
        );
        return Promise.resolve(data);
      })
      .catch((e: Error) =>
        throwError(`Extraction failed for ${data.filepath}: ${e}`)
      )
  );
}


export function getBinaryEntry(
  platform: string = process.platform,
  arch: string = process.arch
): BinaryEntry {
  const found = CONFIG.binaries.find(
    (value) =>
      value.platform === platform && (value.arch ? value.arch === arch : true)
  );
  if (found === undefined) {
    throw makeError(
      `Cannot find binary for platform '${platform}' with architecture '${arch}'.`
    );
  }
  return found;
}


function setup(platform?: string, arch?: string): Promise<Data> {
  const entry = getBinaryEntry(platform, arch);
  return Promise.resolve({
    binaryDownloadPath: join(entry.downloadLocation, entry.binary),
    binaryFilename: entry.binaryFilename,
    checksumDownloadPath: join(PACT_DEFAULT_LOCATION, entry.binaryChecksum),
    filepath: path.resolve(__dirname, entry.binary),
    checksumFilepath: path.resolve(__dirname, entry.binaryChecksum),
    isWindows: pactEnvironment.isWindows(platform),
    platform: entry.platform,
    arch: entry.arch,
    platformFolderPath: path.resolve(__dirname, entry.folderName),
  });
}

// This function is used in the pretest script
export function downloadChecksums(): Promise<void> {
  console.log(chalk.gray(`Downloading All Pact Standalone Binary Checksums.`));
  return Promise.all(
    CONFIG.binaries.map(value =>
      setup(value.platform, value.arch).then((data: Data) =>
        downloadFileRetry(
          data.checksumDownloadPath,
          data.checksumFilepath
        ).then(
          () => {
            console.log(
              chalk.green(
                `Finished downloading checksum ${path.basename(
                  data.checksumFilepath
                )}`
              )
            );
            return data;
          },
          (e: string) =>
            throwError(
              `Error downloading checksum from ${data.checksumDownloadPath}: ${e}`
            )
        )
      )
    )
  ).then(
    () => console.log(chalk.green('All checksums downloaded.')),
    (e: string) => throwError(`Checksum Download Failed Unexpectedly: ${e}`)
  );
}

const HEADER_FILES =['pact.h','pact-cpp.h']

export function downloadHeaders(): Promise<void> {
  console.log(chalk.gray(`Downloading header file.`));
  return Promise.all(
    HEADER_FILES.map(header =>
        downloadFileRetry(
          join(PACT_DEFAULT_LOCATION, header),
          path.resolve(__dirname, header)
        ).then(
          () => {
            console.log(
              chalk.green(
                `Finished downloading checksum ${path.basename(
                  header
                )}`
              )
            );
          },
          (e: string) =>
            throwError(
              `Error downloading header from ${join(PACT_DEFAULT_LOCATION, header)}: ${e}`
            )
        )
      )
  ).then(
    () => console.log(chalk.green('All header downloaded.')),
    (e: string) => throwError(`header Download Failed Unexpectedly: ${e}`)
  );
}


export default (platform?: string, arch?: string): Promise<Data> => {
  if (process.env['PACT_SKIP_BINARY_INSTALL'] === 'true') {
    console.log(
      chalk.yellow(
        "Skipping binary installation. Env var 'PACT_SKIP_BINARY_INSTALL' was found."
      )
    );
    return Promise.resolve({
      binaryInstallSkipped: true,
    } as Data);
  }
  return setup(platform, arch)
    .then(download)
    .then(extract)
    .then(d => {
      console.log(chalk.green('Pact Standalone Binary is ready.'));
      return { ...d, binaryInstallSkipped: false };
    })
    .catch((e: string) =>
      throwError(`Postinstalled Failed Unexpectedly: ${e}`)
    );
};

export interface PackageConfig {
  binaryLocation?: string;
  doNotTrack?: boolean;
}

export interface Config {
  doNotTrack: boolean;
  binaries: BinaryEntry[];
}

export interface Data {
  binaryDownloadPath: string;
  binaryFilename: string;
  checksumDownloadPath: string;
  filepath: string;
  checksumFilepath: string;
  platform: string;
  isWindows: boolean;
  arch?: string;
  platformFolderPath: string;
  binaryInstallSkipped?: boolean;
  binaryAlreadyDownloaded?: boolean;
  binaryAlreadyInstalled?: boolean;
}

export interface BinaryEntry {
  platform: string;
  arch?: string;
  binary: string;
  binaryFilename: string;
  binaryChecksum: string;
  downloadLocation: string;
  folderName: string;
}
